/*
 * arpping.c
 *
 * Mostly stolen from: dhcpcd - DHCP client daemon
 * by Yoichi Hariguchi <yoichi@fore.com>
 *
 * Code Modified by Thomson. Revision 1.0.0
 */



#include "dnsmasq.h"
#include "arpping.h"

/* local prototypes */
static int arpCheck(u_long inaddr, struct ifinfo *ifbuf, long timeout);
static void mkArpMsg(int opcode, u_long tInaddr, u_char *tHaddr, u_long sInaddr, u_char *sHaddr, struct arpMsg *msg);
static int openRawSocket (int *s, u_short type);


/* args:	yiaddr - what IP to ping (eg. on the NETtel cb189701)
 * retn: 	1 addr free
 *		0 addr used
 *		-1 error 
 */  
int arpping(u_int32_t yiaddr, u_int32_t ip, char *interface) {
	struct ifinfo ifbuf;

	struct ifreq ifr;
	int fd;

	/* first get hw addr for intf */
 	if((fd = socket(AF_INET, SOCK_RAW, IPPROTO_RAW)) >= 0) {
	  ifr.ifr_addr.sa_family = AF_INET;
	  strcpy(ifr.ifr_name, interface );

	  if (ioctl(fd, SIOCGIFHWADDR, &ifr) == 0) {
	    memcpy(ifbuf.haddr , ifr.ifr_hwaddr.sa_data, 6);
	    dnsmasq_syslog(LOG_WARNING, "Using intf %s (%02x:%02x:%02x:%02x:%02x:%02x)",
			   interface,
			   ifbuf.haddr[0], ifbuf.haddr[1], ifbuf.haddr[2],
			   ifbuf.haddr[3], ifbuf.haddr[4], ifbuf.haddr[5]);
	  } else {
	    dnsmasq_syslog(LOG_ERR, "SIOCGIFHWADDR failed!");
	    dnsmasq_syslog(LOG_ERR, "Unable to get hw addr, assume addr is free !");
	    return 1;
	  }
  	} else {
	  dnsmasq_syslog(LOG_ERR, "socket failed!");
	  dnsmasq_syslog(LOG_ERR, "Unable to create socket to get hw addr, assume addr is free !");
	  return 1;
	} 
	close(fd);

	strcpy(ifbuf.ifname, interface);
	ifbuf.addr = ip;
	ifbuf.mask = 0x0;
	ifbuf.bcast = 0x0;
	
	ifbuf.flags = 0;
	
	return arpCheck(yiaddr, &ifbuf, 2);
}


static int arpCheck(u_long inaddr, struct ifinfo *ifbuf, long timeout)  {
	int				s;			/* socket */
	int				rv;			/* return value */
	struct sockaddr addr;		/* for interface name */
	struct arpMsg	arp;
	fd_set			fdset;
	struct timeval	tm;
	time_t			prevTime;

	rv = 1;
	openRawSocket(&s, ETH_P_ARP);

	/* send arp request */
	mkArpMsg(ARPOP_REQUEST, inaddr, NULL, ifbuf->addr, ifbuf->haddr, &arp);
	bzero(&addr, sizeof(addr));
	strcpy(addr.sa_data, ifbuf->ifname);
	if ( sendto(s, &arp, sizeof(arp), 0, &addr, sizeof(addr)) < 0 ) rv = 0;
	
	/* wait arp reply, and check it */
	tm.tv_usec = 0;
	time(&prevTime);
	while ( timeout > 0 ) {
		FD_ZERO(&fdset);
		FD_SET(s, &fdset);
		tm.tv_sec  = timeout;
		if ( select(s+1, &fdset, (fd_set *)NULL, (fd_set *)NULL, &tm) < 0 ) {
			dnsmasq_syslog(LOG_ERR, "Error on ARPING request: %s", sys_errlist[errno]);
			if (errno != EINTR) rv = 0;
		} else if ( FD_ISSET(s, &fdset) ) {
			if (recv(s, &arp, sizeof(arp), 0) < 0 ) rv = 0;
			if(arp.operation == htons(ARPOP_REPLY) && 
			   bcmp(arp.tHaddr, ifbuf->haddr, 6) == 0 && 
			   *((u_int *)arp.sInaddr) == inaddr ) {
				dnsmasq_syslog(LOG_INFO, "Valid arp reply receved for this address");
				rv = 0;
				break;
			}
		}
		timeout -= time(NULL) - prevTime;
		time(&prevTime);
	}
	close(s);
	dnsmasq_syslog(LOG_INFO, "%salid arp replies for this address", rv ? "No v" : "V");	 
	return rv;
}

static void mkArpMsg(int opcode, u_long tInaddr, u_char *tHaddr,
		 u_long sInaddr, u_char *sHaddr, struct arpMsg *msg) {
	bzero(msg, sizeof(*msg));
//pgo	bcopy(MAC_BCAST_ADDR, msg->ethhdr.h_dest, 6); /* MAC DA */
//pgo	bcopy(sHaddr, msg->ethhdr.h_source, 6);	/* MAC SA */
//pgo	msg->ethhdr.h_proto = htons(ETH_P_ARP);	/* protocol type (Ethernet) */
	bcopy(MAC_BCAST_ADDR, msg->h_dest, 6); /* MAC DA */
	bcopy(sHaddr, msg->h_source, 6);	/* MAC SA */
	msg->h_proto = htons(ETH_P_ARP);	/* protocol type (Ethernet) */
	msg->htype = htons(ARPHRD_ETHER);		/* hardware type */
	msg->ptype = htons(ETH_P_IP);			/* protocol type (ARP message) */
	msg->hlen = 6;							/* hardware address length */
	msg->plen = 4;							/* protocol address length */
	msg->operation = htons(opcode);			/* ARP op code */
//pgo	*((u_int *)msg->sInaddr) = sInaddr;		/* source IP address */
	bcopy((u_char*)&sInaddr, msg->sInaddr, 4) ;
	bcopy(sHaddr, msg->sHaddr, 6);			/* source hardware address */
//pgo	*((u_int *)msg->tInaddr) = tInaddr;		/* target IP address */
	bcopy((u_char*)&tInaddr, msg->tInaddr, 4) ;
	
	if ( opcode == ARPOP_REPLY )
		bcopy(tHaddr, msg->tHaddr, 6);		/* target hardware address */
}


static int openRawSocket (int *s, u_short type) {
	int optval = 1;

	if((*s = socket (AF_INET, SOCK_PACKET, htons (type))) == -1) {
		dnsmasq_syslog(LOG_ERR, "Could not open raw socket");
		return -1;
	}
	
	if(setsockopt (*s, SOL_SOCKET, SO_BROADCAST, &optval, sizeof (optval)) == -1) {
		dnsmasq_syslog(LOG_ERR, "Could not setsocketopt on raw socket");
		return -1;
	}
	return 0;
}

